/*
 * Copyright (c) 1999-2021 Hartmut Seichter
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
#ifndef PW_CORE_MATRIX_HPP
#define PW_CORE_MATRIX_HPP

#include <pw/core/globals.hpp>
#include <pw/core/math.hpp>
#include <pw/core/matrixbase.hpp>

#include <numeric>
#include <array>

namespace pw
{

	template <std::size_t R, std::size_t C, typename T, bool RowMajor = false>
	struct matrix_ : matrixbase_<T, matrix_<R, C, T>>
	{
		T data[R * C];

		using matrixbase_<T, matrix_<R, C, T>>::matrixbase_;

		static constexpr std::size_t rows{R};
		static constexpr std::size_t cols{C};

		static constexpr std::size_t coefficients{R * C};

		using col_type = matrix_<R, 1, T>;
		using row_type = matrix_<1, C, T>;
		using transpose_type = matrix_<C, R, T>;

		matrix_<R, C, T> &operator=(const matrix_<R, C, T> &other)
		{
			for (size_t i = 0; i < other.size(); i++)
				(*this)[i] = other[i];
			return *this;
		}

		matrix_(std::initializer_list<T> args)
		{
			typename std::initializer_list<T>::iterator it = args.begin();
			for (; it != args.end(); it++)
				data[it - args.begin()] = *it;
		}

		template <typename... Arguments>
		matrix_ &set(Arguments... values)
		{
			static_assert(sizeof...(Arguments) == R * C, "Incorrect number of arguments");
			data = {values...};
			return *this;
		}

		inline size_t offset(size_t r, size_t c) const
		{
			return (RowMajor) ? r * C + c : c * R + r;
		}

		inline T &operator()(std::size_t r, std::size_t c)
		{
			return data[offset(r, c)];
		}

		inline const T &operator()(std::size_t r, std::size_t c) const
		{
			return data[offset(r, c)];
		}

		inline const T *ptr() const { return &data[0]; }

		//! set identity
		inline matrix_ &set_identity()
		{
			for (std::size_t r = 0; r < rows; r++)
				for (std::size_t c = 0; c < cols; c++)
					(*this)(r, c) = (c == r) ? T(1) : T(0);
			return *this;
		}

		inline matrix_ &set_uniform(const T& v)
		{
			std::fill(std::begin(data),std::end(data),v);
			return *this;
		}

		template <std::size_t Rs, std::size_t Cs, bool RowMajorSlice = RowMajor>
		auto slice(std::size_t r, std::size_t c) const
		{
			matrix_<Rs, Cs, T, RowMajorSlice> s;
			for (std::size_t ri = 0; ri < Rs; ri++)
				for (std::size_t ci = 0; ci < Cs; ci++)
					s(ri, ci) = (*this)(ri + r, ci + c);
			return s;
		}

		template <std::size_t Rs, std::size_t Cs, bool RowMajorSlice = RowMajor>
		matrix_ &set_slice(const matrix_<Rs, Cs, T, RowMajorSlice> &s,
						   std::size_t r, std::size_t c)
		{
			for (std::size_t ri = 0; ri < Rs; ri++)
				for (std::size_t ci = 0; ci < Cs; ci++)
					(*this)(ri + r, ci + c) = s(ri, ci);
			return *this;
		}

		template <std::size_t Rs, std::size_t Cs, bool RowMajorSlice = RowMajor>
		auto minor(std::size_t r0, std::size_t c0) const
		{
			matrix_<Rs, Cs, T, RowMajorSlice> m;
			size_t r = 0;
			for (size_t ri = 0; ri < R; ri++)
			{
				size_t c = 0;
				if (ri == r0)
					continue;
				for (size_t ci = 0; ci < C; ci++)
				{
					if (ci == c0)
						continue;
					m(r, c) = (*this)(ri, ci);
					c++;
				}
				r++;
			}
			return m;
		}

		T determinant() const
		{
			T det(0);
			for (size_t c = 0; c < C; c++)
				det += ((c % 2 == 0) ? (*this)(0, c) : -(*this)(0, c)) * this->minor<R - 1, C - 1, RowMajor>(0, c).determinant();
			return det;
		}

		auto transposed() const
		{
			transpose_type res;
			for (size_t r = rows; r-- > 0;)
				for (size_t c = cols; c-- > 0;)
					res(c, r) = (*this)(r, c);
			return res;
		}

		auto inverse() const
		{
			T invDet = T(1) / this->determinant();
			matrix_<R, C, T, RowMajor> inv;
			for (int j = 0; j < C; j++)
				for (int i = 0; i < R; i++)
				{
					const T minorDet = this->minor<R - 1, C - 1, RowMajor>(j, i).determinant();
					const T coFactor = ((i + j) % 2 == 1) ? -minorDet : minorDet;
					inv(i, j) = invDet * coFactor;
				}
			return inv;
		}

		inline bool row_major() const
		{
			return RowMajor;
		}

		inline bool square() const { return R == C; }

		inline const matrix_ operator+(const matrix_ &other) const
		{
			matrix_ res(*this);
			for (size_t r = 0; r < R; r++)
				for (size_t c = 0; c < C; c++)
					res(r, c) += other(r, c);
			return res;
		}

		inline const matrix_ operator-(const matrix_ &other) const
		{
			matrix_ res(*this);
			for (size_t r = 0; r < R; r++)
				for (size_t c = 0; c < C; c++)
					res(r, c) -= other(r, c);
			return res;
		}

		auto row(size_t row_) const
		{
			row_type r;
			for (size_t i = 0; i < cols; i++)
				r[i] = (*this)(row_, i);
			return r;
		}

		auto column(size_t col_) const
		{
			col_type c;
			for (size_t i = 0; i < rows; i++)
				c[i] = (*this)(i, col_);
			return c;
		}

		static constexpr auto identity()
		{
			matrix_ res;
			for (std::size_t r = 0; r < rows; r++)
				for (std::size_t c = 0; c < cols; c++)
					res(r, c) = (c == r) ? T(1) : T(0);
			return res;
		}
	};

	template <>
	inline float matrix_<1, 1, float>::determinant() const
	{
		return (*this)(0, 0);
	}

	template <>
	inline double matrix_<1, 1, double>::determinant() const
	{
		return (*this)(0, 0);
	}

	template <typename T, std::size_t R, std::size_t Ca, std::size_t Cb>
	auto operator*(const matrix_<R, Ca, T> &A,
				   const matrix_<R, Cb, T> &B)
	{
		matrix_<R, Cb, T> result;
		result.zero(); // zero the output
		for (size_t r = 0; r < R; r++)
			for (size_t c = 0; c < Cb; c++)
				for (size_t iI = 0; iI < R; iI++)
					result(r, c) += A(r, iI) * B(iI, c); // inner product
		return result;
	}

	//
	// common matrix types
	//

	template <typename T>
	using matrix2x2_ = matrix_<2, 2, T>;
	template <typename T>
	using matrix3x3_ = matrix_<3, 3, T>;
	template <typename T>
	using matrix4x4_ = matrix_<4, 4, T>;

	using matrix2x2f = matrix2x2_<float>;
	using matrix2x2d = matrix2x2_<double>;
	using matrix2x2 = matrix2x2_<real_t>;

	using matrix3x3f = matrix3x3_<float>;
	using matrix3x3d = matrix3x3_<double>;
	using matrix3x3 = matrix3x3_<real_t>;

	using matrix4x4f = matrix4x4_<float>;
	using matrix4x4d = matrix4x4_<double>;
	using matrix4x4 = matrix4x4_<real_t>;
}

#endif
